package pl.webnet.JaTrac;

import java.io.*;

class Camera implements XmlSettable
{
	protected Point origin;
	Point target;
	Vec direction;
	Vec up;
	Vec right;
	double twist;
	double angleX;
	double angleY;
	double aspect;
	int recursionDepth = 1;
	Screen screen;
	Point imagePlaneUL;
	Point imagePlaneUR;
	Point imagePlaneLL;
	double imagePlaneWidth, imagePlaneHeight;
	Vec imagePlaneRight, imagePlaneUp;
	RayTree rayTree;
	String name = "unnamed";
	Spectrum ambientSpectrum, mediumSpectrum;

	public Camera()
	{
	}

	public Camera(Point origin, Point target, Vec up, double angle, double aspect)
	{
		this.origin = new Point(origin);
		this.target = new Point(target);
		this.up = new Vec(up);
		this.aspect = aspect;
		this.angleX = angle;

		initialize();
	}

	public Camera(Point origin, Point[] imgPlane, double twist)
	{
		this.origin = new Point(origin);

		imagePlaneUL = new Point(imgPlane[0]);
		imagePlaneUR = new Point(imgPlane[1]);
		imagePlaneLL = new Point(imgPlane[2]);

		initialize();
	}

	public Object initialize()
	{
		if((imagePlaneUL != null) && (imagePlaneUR != null) && (imagePlaneLL != null))
		{
			right = new Vec(imagePlaneUR, imagePlaneUL);
			up = new Vec(imagePlaneUL, imagePlaneLL);

			imagePlaneWidth = right.length();
			imagePlaneHeight = up.length();

			up.unit();
			right.unit();
		}
		else
		{
			angleY = angleX / aspect;
			up.unit();

			// calculate camera direction vector
			direction = new Vec(target,  origin);

			// compute size of the image plane in world coordinates;
			imagePlaneWidth = direction.length() * Math.tan(angleX * Math.PI / 180.0 / 2) * 2;
			imagePlaneHeight = direction.length() * Math.tan(angleY * Math.PI / 180.0 / 2) * 2;

			direction.unit();

			// compute cross product of direction and up vectors to get a vector pointing right
			right = up.cross(direction);
			right.unit();

			// compute upper left corner of image plane in world coordinates
			imagePlaneUL = new Point(
				target.x + (-imagePlaneWidth * right.x) / 2 + (imagePlaneHeight * up.x) / 2 - origin.x,
				target.y + (-imagePlaneWidth * right.y) / 2 + (imagePlaneHeight * up.y) / 2 - origin.y,
				target.z + (-imagePlaneWidth * right.z) / 2 + (imagePlaneHeight * up.z) / 2 - origin.z);
		}

	
	//	System.out.println("upperLeft: " + imagePlaneUL);
		return this;
	}

	public String[] getParamArray()
	{
		final String params[] = {
			"origin", "origin", "",
			"imageplane00", "imagePlaneUL", "",
			"imageplane10", "imagePlaneUR", "",
			"imageplane01", "imagePlaneLL", "",
			"target", "target", "",
			"up", "up", "",
			"angle", "angleX", "",
			"recursion", "recursionDepth", "",
			"ratio", "aspect", "",
			"ambient", "ambientSpectrum", "",
			"medium", "mediumSpectrum", "",
			"name", "name", ""
		};

		return params;
	}

	public String[] getAttachArray()
	{
		String params[] = {
			"Screen", "attachScreen", ""
		};

		return params;
	}

	public String toString()
	{
		return(ParamHelper.dumpAttrs(this));
	}

	public void recalcParams()
	{
		double pixelApertureX = imagePlaneWidth / screen.getWidth() ;
		double pixelApertureY = imagePlaneHeight / screen.getHeight() ;

//		System.out.println("pixelApertureX: " + pixelApertureX);
//		System.out.println("pixelApertureY: " + pixelApertureY);

		imagePlaneRight = new Vec(right);
		imagePlaneRight.scale(pixelApertureX);

		imagePlaneUp = new Vec(up);
		imagePlaneUp.scale(pixelApertureY);
		
//		System.out.println("imagePlaneRight:" + imagePlaneRight);
//		System.out.println("imagePlaneUp:" + imagePlaneUp);
	}
	
	public void attachScreen(Screen scr)
	{
		screen = scr;
		screen.init();
		initialize();
		recalcParams();
	}
	
/**
 * Generate a primary (camera) ray.
 * @param		x point on x axis that ray has to go through (screen space)
 * @param		y point on y axis that ray has to go through (screen space)
 * @return		primary ray originating at camera's position and going through <code>(x, y)</code>
 */

	public Ray generateRay(double x, double y)
	{

		Vec dir = new Vec(
			imagePlaneUL.x + (x * imagePlaneRight.x) - (y * imagePlaneUp.x),
			imagePlaneUL.y + (x * imagePlaneRight.y) - (y * imagePlaneUp.y),
			imagePlaneUL.z + (x * imagePlaneRight.z) - (y * imagePlaneUp.z)
			);


//		System.out.println("raydir(" + x + ", " + y + "): " + dir);
//		dir.unit();
		
		return new Ray(origin, dir);
	}

	public void render(Scene scene)
	{
		Spectrum lineCache[][] = null;
		Spectrum pixelCache[][] = null;
		int width = screen.getWidth();
		int height = screen.getHeight();
		int AADepth = 0, pSize = 0, maxRays = 0;
		double t = 0;
		PrintStream tPrintStream = null;

		rayTree = new RayTree(scene, recursionDepth);

		// initialize intersection tester with current scene and camera info
		scene.iTester.initialize(scene, this);

		if(screen.antialias)
		{
			// calculate how deeply the antialiasing routine may subdivide pixel
			// with current threshold
			AADepth = (int)Math.abs(Math.ceil(Math.log(1 / screen.AAThreshold) / Math.log(4)));

			pSize = (1 << AADepth);
			maxRays = pSize + 1;

			System.out.println("AA max. subdivision depth: " + AADepth);
			System.out.println("max rays per axis: " + maxRays);

			// allocate ray cache buffers
			lineCache = new Spectrum[width][maxRays];
			pixelCache = new Spectrum[maxRays][maxRays];
		}
		else
		{
			if(screen.tBufferFilename != null)
			{
				try
				{
					System.out.println("Opening t-buffer file for writing: " + screen.tBufferFilename);
					tPrintStream = new PrintStream(new FileOutputStream(screen.tBufferFilename));
					tPrintStream.println("<?xml version=\"1.0\"?>");

				} catch(FileNotFoundException e) {e.printStackTrace(); System.exit(5); }
			}
		}

		for(int j = 0; j < height; j++)
		{
			System.out.print("\rProcessing line: " + (j + 1) + "/" + height);

			for(int i = 0; i < width; i++)
			{
				Pixel pix = new Pixel();
				Spectrum totalSpectrum = null;

				if(screen.antialias)
				{
					// reuse rays from the bottom of previous row
					for(int x = 0; x < maxRays; x++)
						pixelCache[x][0] = lineCache[i][x];

					totalSpectrum = antialiasPixel(screen.AAThreshold, pixelCache, 0, 0, pSize, i + 0.5, j + 0.5, 0.5);

					// make copy of color information from the bottom line of the pixel
					// for reuse in the next line
					for(int x = 0; x < maxRays; x++)
						lineCache[i][x] = pixelCache[x][maxRays - 1];

					// reuse rays on the right side of the pixel
					for(int y = 0; y < maxRays; y++)
						pixelCache[0][y] = pixelCache[maxRays - 1][y];

					// reset the rest of pixel's ray cache.
					// for bigger caches we could think about ray tagging,
					// to avoid doing this step.
					for(int x = 1; x < maxRays; x++)
						for(int y = 0; y < maxRays; y++)
							pixelCache[x][y] = null;

						//			totalSpectrum = totalSpectrum.create(Spectrum.white);

						//totalSpectrum.scale(1 - 1. / pix.AARays );
				}
				else
				{
					// when not using AA, just shoot a ray through pixel's center
					Ray ray = rayTree.trace(0, 1., generateRay(i + 0.5, j + 0.5));
					totalSpectrum = ray.spectrum;

					if(tPrintStream != null)
					{
						if(ray.intersections.isEmpty())
							t = Double.POSITIVE_INFINITY;
						else
						{
							Intersection intersection = (Intersection)ray.intersections.first();

							t = intersection.t / 20;
						}

						tPrintStream.println("<pixel x=\"" + i + "\" y=\"" + j + "\" t=\"" + t + "\" />");
					}
				}
				
				pix.set(totalSpectrum);
				screen.pixels[j][i] = pix;
			}

			// reset pixel's ray cache at the beginning of a new line
			for(int y = 0; y < maxRays; y++)
				pixelCache[0][y] = null;
		}
	}

	protected Spectrum antialiasPixel(double threshold, Spectrum pixelCache[][], int pX, int pY, int pSize, double x, double y, double size)
	{
		Spectrum sA, sB, sC, sD;
		
		// send rays through corners of the (sub)pixel) if they
		// are not already in the cache

		if(pixelCache[pX][pY] == null)
			pixelCache[pX][pY] = rayTree.trace(0, 1., generateRay(x - size, y - size)).spectrum;
		
		if(pixelCache[pX + pSize][pY] == null)
			pixelCache[pX + pSize][pY] = rayTree.trace(0, 1., generateRay(x + size, y - size)).spectrum;

		if(pixelCache[pX + pSize][pY + pSize] == null)
			pixelCache[pX + pSize][pY + pSize] = rayTree.trace(0, 1., generateRay(x + size, y + size)).spectrum;

		if(pixelCache[pX][pY + pSize] == null)
			pixelCache[pX][pY + pSize] = rayTree.trace(0, 1., generateRay(x - size, y + size)).spectrum;

		// send ray through the center of (sub)pixel
		Spectrum center = rayTree.trace(0, 1., generateRay(x, y)).spectrum;
		Spectrum result = center.create(Spectrum.black);

		if(threshold > 1)
		{
			result.add(pixelCache[pX][pY]);
			result.add(pixelCache[pX + pSize][pY]);
			result.add(pixelCache[pX + pSize][pY + pSize]);
			result.add(pixelCache[pX][pY + pSize]);
			result.scale(0.25);	// 1/4

			result.add(center);
			result.scale(0.5);

			return result;
		}

		double halfSize = size / 2;
		int halfPSize = pSize / 2;

		pixelCache[pX + halfPSize][pY + halfPSize] = center;

		// upper-left
		if(center.difference(pixelCache[pX][pY]) > threshold)
			sA = antialiasPixel(threshold * 4, pixelCache, pX, pY, halfPSize, x - halfSize, y - halfSize, halfSize);
		else
		{
			sA = center.create(pixelCache[pX][pY]);
			sA.average(center);
		}
 
		// upper-right
		if(center.difference(pixelCache[pX + pSize][pY]) > threshold)
			sB = antialiasPixel(threshold * 4, pixelCache, pX + halfPSize, pY, halfPSize, x + halfSize, y - halfSize, halfSize);
		else
		{
			sB = center.create(pixelCache[pX + pSize][pY]);
			sB.average(center);
		}
 
		// bottom-right
		if(center.difference(pixelCache[pX + pSize][pY + pSize]) > threshold)
			sC = antialiasPixel(threshold * 4, pixelCache, pX + halfPSize, pY + halfPSize, halfPSize, x + halfSize, y + halfSize, halfSize);
		else
		{
			sC = center.create(pixelCache[pX + pSize][pY + pSize]);
			sC.average(center);
		}
 
		// bottom-left
		if(center.difference(pixelCache[pX][pY + pSize]) > threshold)
			sD = antialiasPixel(threshold * 4, pixelCache, pX, pY + halfPSize, halfPSize, x - halfSize, y + halfSize, halfSize);
		else
		{
			sD = center.create(pixelCache[pX][pY + pSize]);
			sD.average(center);
		}
 
		// average four corners
		result.add(sA);
		result.add(sB);
		result.add(sC);
		result.add(sD);

		result.scale(0.25);	// 1/4

		return result;
	}
/*	
	protected Spectrum antialiasPixel(Pixel pix, double weight, double x, double y, double size, Spectrum sA, Spectrum sB, Spectrum sC, Spectrum sD)
	{
		double threshold = weight;
		Spectrum sAB, sBC, sCD, sDA;

//		pix.AARays++;

		sAB = sBC = sCD = sDA = null;

		// clone spectral values to avoid modifying the originals
		try {
			sA = (Spectrum)sA.clone();
			sB = (Spectrum)sB.clone();
			sC = (Spectrum)sC.clone();
			sD = (Spectrum)sD.clone();
		} catch (CloneNotSupportedException e) {}


	//	System.out.println("Entering antialiasPixel: weight=" + weight);

		// shoot ray through region's center
		Spectrum sE = rayTree.trace(0, 1., generateRay(x, y)).spectrum;

		// no difference can be bigger than this.
		if(weight > 1)
		{
			sA.add(sB);
			sA.add(sC);
			sA.add(sD);

			sA.scale(0.25);	// 1/4
			sA.add(sE);
			sA.scale(0.5);

			return sA;
		}

		double halfSize = size / 2;
		

		// check if differences between corners and center are large enough
		// to subdivide further.

		// upper-left
		if(sA.difference(sE) > threshold)
			sA = antialiasPixel(pix, weight * 4, x - halfSize, y - halfSize, halfSize,
				sA,
				sAB = rayTree.trace(0, 1., generateRay(x, y - size)).spectrum,
				sE,
				sDA = rayTree.trace(0, 1., generateRay(x - size, y)).spectrum);
		else
			sA.average(sE);

		// upper-right
		if(sB.difference(sE) > threshold)
			sB = antialiasPixel(pix, weight * 4, x + halfSize, y - halfSize, halfSize,
				sAB != null ? sAB : rayTree.trace(0, 1., generateRay(x, y - size)).spectrum,
				sB,
				sBC = rayTree.trace(0, 1., generateRay(x + size, y)).spectrum,
				sE);
		else
			sB.average(sE);

		// bottom-right
		if(sC.difference(sE) > threshold)
			sC = antialiasPixel(pix, weight * 4, x + halfSize, y + halfSize, halfSize,
				sE,
				sBC != null ? sBC : rayTree.trace(0, 1., generateRay(x + size, y)).spectrum,
				sC,
				sCD = rayTree.trace(0, 1., generateRay(x, y + size)).spectrum);
		else
			sC.average(sE);

		// bottom-left
		if(sD.difference(sE) > threshold)
			sD = antialiasPixel(pix, weight * 4, x - halfSize, y + halfSize, halfSize,
				sDA != null ? sDA : rayTree.trace(0, 1., generateRay(x - size, y)).spectrum,
				sE,
				sCD != null ? sCD : rayTree.trace(0, 1., generateRay(x, y + size)).spectrum,
				sD);
		else
			sD.average(sE);


		// average four corners
		sA.add(sB);
		sA.add(sC);
		sA.add(sD);

		sA.scale(0.25);	// 1/4

		return sA;
	}
*/
}
